/*
 * read.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * Reads the entire file into memory.
 * Allocates the required amount of memory.
 * The provided buffer must be not allocated yet.
 */
void read_entire_file(char **buffer,unsigned long long *buffer_size,char *filename)
{
    FILE *f;
    unsigned long long filesize;

    assert(buffer);
    assert(buffer_size);
    assert(!*buffer);
    assert(!*buffer_size);
    assert(filename);

    f = fopen(filename,"rb");
    if (!f) { fprintf(stderr,"Can't open file \"%s\"\n",filename); exit(1); }

    fseek(f,0,SEEK_END);
    filesize = ftell(f);
    fseek(f,0,SEEK_SET);

    *buffer = (char*)malloc1(filesize+1);
    *buffer_size = fread(*buffer,1,filesize,f);
    fclose(f);
    (*buffer)[filesize] = 0;

    if (*buffer_size != filesize) { fprintf(stderr,"Can't read %llu bytes from input file\n",filesize); exit(1); }
}

/*
 * Parses sequences in FASTA format.
 * Allocates and populates the seqs array.
 */
void parse_fasta(char *fasta,unsigned long long fasta_size)
{
    char *end = fasta + fasta_size, *c;
    unsigned long long i = 0, line = 1;
    seq_t *seq;

    n_seqs = 0ull;
    for (c=fasta; c<end; c++) { if (*c == '>') { n_seqs++; } }
    if (!n_seqs) { fprintf(stderr,"Input has no sequences\n"); exit(1); }

    seqs = (seq_t*)malloc1(seq_size*n_seqs);
    for (c=fasta,seq=seqs; ; seq=(seq_t*)(((void*)seq)+seq_size))
    {
        if (*c != '>') { fprintf(stderr,"Can't find sequence start ('>') in line %llu\n",line); exit(1); }
        seq->name = c;
        do { c++; } while (c<end && *c!=13 && *c!=10);
        if (c>=end || (*c!=13 && *c!=10)) { fprintf(stderr,"Can't find end of sequence name\n"); exit(1); }
        while (c<end && (*c==13 || *c==10)) { *c = 0; c++; }
        line++;
        if (c>=end) { fprintf(stderr,"Can't find data for sequence \"%s\"\n",seq->name); exit(1); }
        seq->data = c;
        i++;
        while (c<end && *c!=13 && *c!=10) { c++; }
        if (c>=end) { break; }
        while (c<end && (*c==13 || *c==10 || *c==0)) { *c = 0; c++; }
        line++;
        if (c>=end) { break; }
    }
    if (i != n_seqs) { fprintf(stderr,"Failed to parse %llu sequences\n",n_seqs); exit(1); }
    fprintf(stderr,"Parsed %llu sequences\n",i);
    /*fprintf(stderr,"Allocated %llu bytes (%llu + %llu)\n",input_size+1+seq_size*n_seqs,input_size+1,seq_size*n_seqs);*/
}

/*
 * Parses barcode sequence. Extracts variable positions of barcode and stores them in binary form.
 * if barcode does not match template, or is too short, or contains unknown characters, the first character of sequence name will be changed to '#'.
 * Returns:
 *    0 - barcode was read successfully.
 *    1 - barcode does not match template (including when barcode is shorter than template).
 *    2 - barcode has characters other than [ACGT].
 */
unsigned convert_barcode_to_binary(seq_t* seq)
{
    unit_t *u = (unit_t*)(seq+1);
    unsigned u_filled = 0, u_filled_total = 0, t_pos = 0;
    unsigned char *c = (unsigned char*)(seq->data);

    *u = 0ull;
    /*printf("Converting \"%s\" to binary\n",seq->data);*/
    for (; *c && t_pos<template_length; c++,t_pos++)
    {
        /*printf("  %c",*c);*/
        unsigned nuc = is_nucleotide[*c];
        /*printf(" -> %d",nuc);*/
        if (!nuc) { *(seq->name) = '#'; /*printf("Unknown character \"%c\" in barcode!\n",*c);*/ return 2; }
        if (use_barcode_position[t_pos])
        {
            /*printf(" (use)");*/
            if (u_filled >= BP_PER_UNIT) { *++u = 0ull; u_filled = 0; }
            *u = (*u << 2) | (nuc-1);
            u_filled++;
            u_filled_total++;
        }
        else
        {
            /*printf(" (=%c=)",template[t_pos]);*/
            if (toupper(*c) != template[t_pos]) { *(seq->name) = '#'; /*printf(" - mismatch!\n");*/ return 1; }
        }
        /*printf(" (filled %u)\n",u_filled_total);*/
    }
    /*if (*c) { *(seq->name) = '#'; printf("  - barcode has extra characters\n"); return 1; }*/
    if (u_filled_total != barcode_complexity) { *(seq->name) = '#'; /*printf("  - loaded %u characters, need %u\n",u_filled_total,barcode_complexity);*/ return 1; }
    *u <<= (BP_PER_UNIT - u_filled)*2;
    return 0;
}

/*
 * Goes through all sequences, extracts variable positions of barcode from sequence and stores them in binary form.
 * Annotates failures with '#' as the first character of sequence name (instead of '>').
 */
void convert_all_barcodes_to_binary()
{
    unsigned long long n_parse_errors = 0ull, n_template_mismatches = 0ull, n_ok = 0ull, i;

    assert(seqs);
    assert(n_seqs);

    fprintf(stderr,"Converting barcodes to binary form\n");

    for (i=0ull; i<n_seqs; i++)
    {
        seq_t *seq = (seq_t*)(((void*)seqs)+seq_size*i);
        unsigned error = convert_barcode_to_binary(seq);
        if (!error) { n_ok++; }
        else if (error == 1) { n_template_mismatches++; }
        else { n_parse_errors++; }
    }

    if (n_parse_errors) { fprintf(stderr," - %llu parse errors\n",n_parse_errors); }
    if (n_template_mismatches) { fprintf(stderr," - %llu template mismatches\n",n_template_mismatches); }
    fprintf(stderr," - %llu barcodes match template\n",n_ok);
    if (!n_ok) { fprintf(stderr,"No barcodes match template\n"); exit(1); }
    n_seqs_ok = n_ok;
}

#endif
